\\ Coppersmith--Howgrave-Graham certificate tester, Version 0.8
\\ David Broadhurst, 3 Mar 2006, with huge help from John Renze
\\ (now allowing Master option and John's Cartesian improvement)

\\ Format of cert : [Id,n,F,G,U,V] with Id=string,
\\ U[k]=[h,u,q,Y,pol(x)] where Y increases with k,
\\ witness q>1 shows that pol has no integer root,
\\ if min(F,G)>1, V is like U, else it is ignored.
\\ Master option with min(F,G)=1: if pol=0 then 
\\ pol is computed by calling Master(u,max(F,G),n).

{CHGcertD(C)=                                                 \\ C=[Id,n,F,G,U,V]
local(Id,n,F,G,nj,s,Ym,U,m,X,K,h,u,q,Y,pol,a,hu,b,c,T,kj,n1); \\ nj is #classes
Id=C[1];print("\nTesting a PRP called \""Id"\".\n");          \\ print target
gettime;n=C[2];F=C[3];G=C[4];nj=1+(min(F,G)>1);s=lcm(F,G);    \\ mend J's chinese
if(notN([(n-1)/2,F,G])||(n-1)%F||(n+1)%G,abend("Modulus"));   \\ check modulus
if(gcd(F,G)>1,if(F%4==2,F=F/2,G=G/2));                        \\ mend J's mod 4
if(n<max(max(F,G)^(10/3),max(s,10^333)^3),abend("Peanuts"));  \\ reject peanuts
if(s<n^0.251,abend("Joker"));                                 \\ reject jokes
Ym=min(if(F>G,sqtest1(n,F,G),sqtest2(n,F,G)),nomod(n,s));     \\ get endpoint
for(j=1,nj,U=C[4+j];m=lift(if(j<2,1,n%s)/Mod(s,n));X=1;       \\ class loop
for(k=1,length(U),K=U[k];h=K[1];u=K[2];q=K[3];Y=K[4];pol=K[5];\\ U=[h,u,q,Y,pol]
if(pol,,if(min(F,G)>1,abend("Alg"),pol=Master(u,max(F,G),n)));\\ Master option
a=pol;if(notN([q-1,h,u,h-u,Y,subst(a,x,1)^2]),abend("Type")); \\ check numbers
hu=[h,u];if(poldegree(pol)!=h-1,abend("Degree"));             \\ check degree
forstep(i=h-1,0,-1,b=if(i>u,x^(i-u)*(x+m)^u,n^(u-i)*(x+m)^i); \\ use CHG basis
c=pollead(a)/pollead(b);a=a-b*c;if(notZ(c),abend("LLL")));    \\ check reduction
T=h*norml2(vector(h,i,polcoeff(pol,i-1)*X^(i-1)))/(X*s)^(2*u);\\ mend J's indices
kj=[k,j];n1=precision(sqrt(T),1);if(T>1,abend("Norm"));       \\ test at X 
print1("Pol"kj" with [h, u]="hu" has ratio="n1" at X, ");     \\ ratio at X
T=h*norml2(vector(h,i,polcoeff(pol,i-1)*Y^(i-1)))/(Y*s)^(2*u);\\ test at Y
if(T>1,abend("Norm"),print1("ratio="precision(sqrt(T),1)));   \\ ratio at Y 
for(i=1,q,if(subst(pol,x,i)%q==0,abend("Perjury")));          \\ check witness
print(" at Y, witness="q".");X=Y+1);if(Y<Ym,abend("Ymax")));  \\ check endpoint
print("\nValidated in "ceil(gettime/10^3)" sec.\n")}          \\ print time

nomod(n,s)=ceil(sqrtint(n)/s)                                 \\ safe ceiling

{sqtest1(n,F,G)=local(r,R,a,b);                               \\ F >> G
r=((n-1)/F)%G;R=(n-1-F*r)/(F*G);a=floor(R/F);b=R-a*F;         \\ D's method
if(issquare((G*b+r)^2-4*G*a),abend("Issquare1"));             \\ test for factor
ceil(2*n/(F^3*G^2))}                                          \\ safe ceiling

{sqtest2(n,F,G)=local(r,R,a,b,c);                             \\ G >> F
r=((n+1)/G)%F;R=(n+1-G*r)/(F*G);a=floor(R/G);b=R-a*G;c=G-b;   \\ mend J's floor
if(issquare((F*b+r)^2+4*F*a),abend("Issquare2_1"));           \\ mend J's sign
if(issquare((F*c-r)^2+4*F*(a+1)),abend("Issquare2_2"));       \\ mend J's logic
ceil(n/(F^2*G^3))}                                            \\ safe ceiling

notZ(n)=type(n)!="t_INT"                                      \\ not an integer

{notN(ls)=local(n);                                           \\ not numbers
for(k=1,length(ls),n=ls[k];if(notZ(n)||n<1,return(1)));0}     \\ scan the list

abend(msg)=print("\""msg"\" error, so we quit!");quit()       \\ print error
